#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<time.h>
#include<string.h>
#include<cuda.h>
#include<cufft.h>
#include"./parameters.h"

#define RSIZE (NZ*NY*NX*sizeof(cufftReal))
#define CSIZE (NZ*NY*PN*sizeof(cufftComplex))

#define FORWARD 0
#define INVERSE 1

#define NTHREADS 1024

#define DIRX 0
#define DIRY 1
#define DIRZ 2

#define PN (NX/2+1)
#define KMAX ((PI/HH)*2.0f/3.0f)
#define PI 3.14159265f

// *** Define GPU variables ***
__device__ cufftReal *rho_r;
__device__ cufftReal *fnl_r;
__device__ cufftComplex *rho_k;
__device__ cufftComplex *fnl_k;
__device__ cufftComplex *pde_k;
__device__ cufftComplex *rhs_k;
cufftHandle fft_forward,fft_inverse;

// *** Define function prototypes ***
int assemble_solver(int);
int destroy_solver(void);
int fft(cufftReal *real_array,cufftComplex *complex_array,int dir);
int copy_cpu2gpu(float *,cufftReal *);
int copy_rhor_gpu2cpu(cufftReal *,float *);
int copy_rhok_gpu2cpu(cufftComplex *,float2 *);
int calc_nonlin(float *,float *);
int calc_pde(float2 *,float2 *,float2 *);
int multiply(float *,float,float);
int partial(float2 *,int);
int partial_add(float2 *,float2 *,int);
int update_rhok(float2 *,float2 *);
int init_rhs(float2 *);
int convolution(void);
int fdensity(void);
int initialize_rho(float *);
int save_data(float *,char *);
int load_data(float *,char *);
int save_rhor_XYZ(float *,char *);
int save_rhok_XYZ(float2 *,char *);
int cut_data(float *,float *,float,float);
float compare(float *,float *);

// *** Define kernel prototypes ***
__global__ void norm_kernel(cufftComplex *,int,float);
__global__ void filter_kernel(cufftComplex *,int,float);
__global__ void nonlin_kernel(float *,float *,int);
__global__ void update_kernel(float2 *,float2 *,int);
__global__ void convolution_kernel(float2 *,float2 *,int);
__global__ void fdensity_kernel(float *,float *,int);
__global__ void partial_kernel(float2 *,int,int);
__global__ void null_kernel(float2 *,int);
__global__ void partial_add_kernel(float2 *,float2 *,int,int);
__global__ void pde_kernel(float2 *,float2 *,float2 *,int);
__global__ void multiply_kernel(float *,float,float,int);
__global__ void compare_kernel(float *,float *,int);

main()
{
  char fname[100];
  float dist;

  // *** Initialize GPU device ***
  assemble_solver(DEVICE);

  // *** Allocate CPU array ***
  float *cpu_data,*cpu_original;
  cpu_data=(float *)malloc(RSIZE);
  cpu_original=(float *)malloc(RSIZE);
  float2 *cpu_spectrum;
  cpu_spectrum=(float2 *)malloc(CSIZE);

  float fbar,pbar=0.0f;
  FILE *fid;

  //initialize_rho(cpu_data);
  load_data(cpu_original,"start.dat");
  save_rhor_XYZ(cpu_original,"reference");

  cut_data(cpu_original,cpu_data,R0,D0);
  copy_cpu2gpu(cpu_data,rho_r);
  fft(rho_r,rho_k,FORWARD); // rho_r -> rho_k
  cudaMemcpy(rho_k,&pbar,sizeof(float),cudaMemcpyHostToDevice);

  fid=fopen("time.dat","wt");
  fclose(fid);

  int s=0;
  for (int n=0; n<NSTEPS; n++)
  {
    cudaMemcpy(fnl_k,rho_k,CSIZE,cudaMemcpyDeviceToDevice); // make a copy of rho_k to fnl_k
    fft(rho_r,fnl_k,INVERSE); // fnl_k -> rho_r
    calc_nonlin(rho_r,fnl_r); // rho_r -> fnl_r
    fft(fnl_r,fnl_k,FORWARD); // fnl_r -> pde_k
    calc_pde(rho_k,fnl_k,pde_k); // fnl_k & rho_k -> pde_k;

    init_rhs(rhs_k);
    for (int dir=0; dir<3; dir++)
    {
      cudaMemcpy(fnl_k,pde_k,CSIZE,cudaMemcpyDeviceToDevice); // make a copy of rho_k to fnl_k
      partial(fnl_k,dir);
      fft(fnl_r,fnl_k,INVERSE);
      multiply(fnl_r,R0,D0);
      fft(fnl_r,fnl_k,FORWARD);
      partial_add(rhs_k,fnl_k,dir);
   }

    update_rhok(rho_k,rhs_k); // rho_k & fnl_k -> rho_k

    if (!(n%SSTEPS))
    {
      sprintf(fname,"spectrum_%03d",s);
      copy_rhok_gpu2cpu(rho_k,cpu_spectrum);
      save_rhok_XYZ(cpu_spectrum,fname);
      cudaMemcpy(fnl_k,rho_k,CSIZE,cudaMemcpyDeviceToDevice); // make a copy of rho_k to fnl_k
      fft(rho_r,fnl_k,INVERSE); // fnl_k -> rho_r
      sprintf(fname,"signal_%03d",s);
      copy_rhor_gpu2cpu(rho_r,cpu_data);
      save_rhor_XYZ(cpu_data,fname);
      save_data(cpu_data,"result.dat");
      s++;
    }

    if ((!(n%MSTEPS))&&n)
    {
      cudaMemcpy(&pbar,rho_k,sizeof(float),cudaMemcpyDeviceToHost);
      pbar/=sqrtf(NZ*NY*NX);
      convolution(); // rho_k -> fnl_k
      fft(fnl_r,fnl_k,INVERSE); // fnl_k -> fnl_r
      cudaMemcpy(fnl_k,rho_k,CSIZE,cudaMemcpyDeviceToDevice); // make a copy of rho_k to fnl_k
      fft(rho_r,fnl_k,INVERSE); // fnl_k -> rho_r
      fdensity(); // fnl_r & rho_r -> fnl_r
      fft(fnl_r,fnl_k,FORWARD); // fnl_r -> fnl_k
      cudaMemcpy(&fbar,fnl_k,sizeof(float),cudaMemcpyDeviceToHost);
      fbar/=sqrtf(NZ*NY*NX);
      dist=sqrtf(3.0f/(4.0f*M_PI)*powf(HH/R0,3.0f)*compare(rho_r,cpu_original));
      printf("%15.10e %15.10e %15.10e\n",n*DT,pbar,fbar);
      fid=fopen("time.dat","at");
      fprintf(fid,"%15.10e %15.10e %15.10e %15.10e\n",n*DT,pbar,fbar,dist);
      fclose(fid);
    }
  }

  copy_rhok_gpu2cpu(rho_k,cpu_spectrum);
  save_rhok_XYZ(cpu_spectrum,"spectrum_final");
  cudaMemcpy(fnl_k,rho_k,CSIZE,cudaMemcpyDeviceToDevice); // make a copy of rho_k to fnl_k
  fft(rho_r,fnl_k,INVERSE); // fnl_k -> rho_r
  copy_rhor_gpu2cpu(rho_r,cpu_data);
  save_rhor_XYZ(cpu_data,"signal_final");
  save_data(cpu_data,"result.dat");

  // *** Free CPU array ***
  free(cpu_spectrum);
  free(cpu_data);
  free(cpu_original);

  // *** Reset GPU device ***
  destroy_solver();
}

// *** USER DEFINED FUNCTIONS

int initialize_rho(float *data)
{
  float f=1.0f/sqrtf(2.0f);

  time_t tp=time(NULL);
  unsigned long int seed = tp % 173683;
  srand(seed);

  for (int k=0; k<NZ; k++)
    for (int j=0; j<NY; j++)
      for (int i=0; i<NX; i++)
      {
        int idx=(k*NY+j)*NX+i;
        float x=(i-NX/2)*HH;
        float y=(j-NY/2)*HH;
        float z=(k-NZ/2)*HH;
        float r=sqrtf(x*x+y*y+z*z);
        float fbcc=(cosf(x*f)*cosf(y*f)+cosf(x*f)*cosf(z*f)+cosf(y*f)*cosf(z*f))/3.0f;
        data[idx]=BCCA*fbcc+RAMP*(2.0f*rand()/RAND_MAX-1.0f);
      }

  return(1);
}

int save_data(float *data,char *fname)
{
  FILE *fid;

  fid=fopen(fname,"wt");
  for (int k=0; k<NZ; k++)
    for (int j=0; j<NY; j++)
      for (int i=0; i<NX; i++)
      {
        int idx=(k*NY+j)*NX+i;
        fprintf(fid,"%15.10e ",data[idx]);
      }
  fclose(fid);

  return(1);
}

int load_data(float *data,char *fname)
{
  FILE *fid;

  fid=fopen(fname,"rt");
  for (int k=0; k<NZ; k++)
    for (int j=0; j<NY; j++)
      for (int i=0; i<NX; i++)
      {
        int idx=(k*NY+j)*NX+i;
        fscanf(fid,"%f",&(data[idx]));
      }
  fclose(fid);

  return(1);
}

int cut_data(float *src,float *dst,float r0,float d0)
{
  float x,y,z,r,f;

  for (int k=0; k<NZ; k++)
  {
    z=(k-NZ/2)*HH;
    for (int j=0; j<NY; j++)
    {
      y=(j-NY/2)*HH;
      for (int i=0; i<NX; i++)
      {
        x=(i-NX/2)*HH;
        r=sqrt(x*x+y*y+z*z);
        f=1.0f-MOBFNC(r,r0,d0);
        int idx=(k*NY+j)*NX+i;
        dst[idx]=f*src[idx];
      }
    }
  }

  return(1);
}

int save_rhor_XYZ(float *data,char *str)
{
  FILE *fid;
  char fname[100];

  strcpy(fname,str);
  strcat(fname,"_XY.dat");
  fid=fopen(fname,"wt");
  for (int j=0; j<NY; j++)
  {
    for (int i=0; i<NX; i++)
    {
      int idx=((NZ/2)*NY+j)*NX+i;
      fprintf(fid,"%15.10e ",data[idx]);
    }
    fprintf(fid,"\n");
  }
  fclose(fid);

  strcpy(fname,str);
  strcat(fname,"_XZ.dat");
  fid=fopen(fname,"wt");
  for (int k=0; k<NZ; k++)
  {
    for (int i=0; i<NX; i++)
    {
      int idx=(k*NY+(NY/2))*NX+i;
      fprintf(fid,"%15.10e ",data[idx]);
    }
    fprintf(fid,"\n");
  }
  fclose(fid);

  strcpy(fname,str);
  strcat(fname,"_YZ.dat");
  fid=fopen(fname,"wt");
  for (int k=0; k<NZ; k++)
  {
    for (int j=0; j<NY; j++)
    {
      int idx=(k*NY+j)*NX+(NX/2);
      fprintf(fid,"%15.10e ",data[idx]);
    }
    fprintf(fid,"\n");
  }
  fclose(fid);

  return(1);
}

int2 convert(int3 coords)
{
  int flag=1;
  if (coords.x<0)
    flag=-1;

  int bk=coords.z*flag;
  int bj=coords.y*flag;
  int bi=coords.x*flag;

  int k=( bk<0 ? bk+NZ : bk );
  int j=( bj<0 ? bj+NY : bj );
  int i=bi;

  int2 res;
  res.x=(k*NY+j)*PN+i;
  res.y=flag;

  return(res);
}

int save_rhok_XYZ(float2 *data,char *str)
{
  FILE *fid1,*fid2;
  char fname[100];
  int3 coords;
  int2 idx;
  int i,j,k;
  float2 fk;
  float2 zero;
  zero.x=0.0f;
  zero.y=0.0f;

  strcpy(fname,str);
  strcat(fname,"_XY_Re.dat");
  fid1=fopen(fname,"wt");
  strcpy(fname,str);
  strcat(fname,"_XY_Im.dat");
  fid2=fopen(fname,"wt");
  k=0;
  for (j=-NY/2; j<NY/2; j++)
  {
    for (i=-NX/2; i<NX/2; i++)
    {
      coords.x=i;
      coords.y=j;
      coords.z=k;
      idx=convert(coords);
      fk=( idx.x ? data[idx.x] : zero );
      fprintf(fid1,"%15.10e ",fk.x);
      fprintf(fid2,"%15.10e ",idx.y*fk.y);
    }
    fprintf(fid1,"\n");
    fprintf(fid2,"\n");
  }
  fclose(fid1);
  fclose(fid2);

  strcpy(fname,str);
  strcat(fname,"_XZ_Re.dat");
  fid1=fopen(fname,"wt");
  strcpy(fname,str);
  strcat(fname,"_XZ_Im.dat");
  fid2=fopen(fname,"wt");
  j=0;
  for (k=-NZ/2; k<NZ/2; k++)
  {
    for (i=-NX/2; i<NX/2; i++)
    {
      coords.x=i;
      coords.y=j;
      coords.z=k;
      idx=convert(coords);
      fk=( idx.x ? data[idx.x] : zero );
      fprintf(fid1,"%15.10e ",fk.x);
      fprintf(fid2,"%15.10e ",idx.y*fk.y);
    }
    fprintf(fid1,"\n");
    fprintf(fid2,"\n");
  }
  fclose(fid1);
  fclose(fid2);

  strcpy(fname,str);
  strcat(fname,"_YZ_Re.dat");
  fid1=fopen(fname,"wt");
  strcpy(fname,str);
  strcat(fname,"_YZ_Im.dat");
  fid2=fopen(fname,"wt");
  i=0;
  for (k=-NZ/2; k<NZ/2; k++)
  {
    for (j=-NY/2; j<NY/2; j++)
    {
      coords.x=i;
      coords.y=j;
      coords.z=k;
      idx=convert(coords);
      fk=( idx.x ? data[idx.x] : zero );
      fprintf(fid1,"%15.10e ",fk.x);
      fprintf(fid2,"%15.10e ",idx.y*fk.y);
    }
    fprintf(fid1,"\n");
    fprintf(fid2,"\n");
  }
  fclose(fid1);
  fclose(fid2);

  return(1);
}

// *** NEVER TOUCH THIS PART ***

int assemble_solver(int device)
{
  // *** Select CUDA device ***
  cudaSetDevice(device);

  // *** Allocate memory on GPU ***
  cudaMalloc((void**)&rho_r,RSIZE);
  cudaMalloc((void**)&fnl_r,RSIZE);
  cudaMalloc((void**)&rho_k,CSIZE);
  cudaMalloc((void**)&fnl_k,CSIZE);
  cudaMalloc((void**)&pde_k,CSIZE);
  cudaMalloc((void**)&rhs_k,CSIZE);

  // *** Define FFT and inverse FFT plans
  cufftPlan3d(&fft_forward,NZ,NY,NX,CUFFT_R2C);
  cufftPlan3d(&fft_inverse,NZ,NY,NX,CUFFT_C2R);

  return(1);
}

int destroy_solver(void)
{
  // *** Clear cuFFT plans ***
  cufftDestroy(fft_forward);
  cufftDestroy(fft_inverse);

  // *** Free memory on GPU ***
  cudaFree(rho_r);
  cudaFree(fnl_r);
  cudaFree(rho_k);
  cudaFree(fnl_k);
  cudaFree(pde_k);
  cudaFree(rhs_k);

  // *** Reset CUDA device ***
  cudaDeviceReset();

  return(1);
}

int copy_cpu2gpu(float *cpu_data,cufftReal *gpu_data)
{
  cudaMemcpy(gpu_data,cpu_data,RSIZE,cudaMemcpyHostToDevice);

  return(1);
}

int copy_rhor_gpu2cpu(cufftReal *gpu_data,float *cpu_data)
{
  cudaMemcpy(cpu_data,gpu_data,RSIZE,cudaMemcpyDeviceToHost);

  return(1);
}

int copy_rhok_gpu2cpu(cufftComplex *gpu_data,float2 *cpu_data)
{
  cudaMemcpy(cpu_data,gpu_data,CSIZE,cudaMemcpyDeviceToHost);

  return(1);
}

int fft(cufftReal *real_array,cufftComplex *complex_array,int dir)
{
  float factor=1.0f/sqrtf(NZ*NY*NX);

  if (dir==FORWARD)
  {
    cufftExecR2C(fft_forward,real_array,complex_array);
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf((float)NZ*NY*PN/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    filter_kernel<<<grid,threads>>>(complex_array,NZ*NY*PN,factor);
  }
  if (dir==INVERSE)
  {
    cufftExecC2R(fft_inverse,complex_array,real_array);
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf((float)NZ*NY*(NX/2)/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    norm_kernel<<<grid,threads>>>((cufftComplex *)real_array,NZ*NY*NX/2,factor);
  }

  return(1);
}

int calc_nonlin(float *src,float *dst)
{
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*NX/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    nonlin_kernel<<<grid,threads>>>(src,dst,NZ*NY*NX);

    return(1);
}

float compare(float *src,float *ref)
{
    float res=0.0f;

    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*NX/NTHREADS)+1;
    dim3 grid(blocks,blocks);

    cudaMemcpy(fnl_r,ref,NZ*NY*NX*sizeof(float),cudaMemcpyHostToDevice);
    compare_kernel<<<grid,threads>>>(src,fnl_r,NZ*NY*NX);
    fft(fnl_r,fnl_k,FORWARD);
    cudaMemcpy(&res,fnl_k,sizeof(float),cudaMemcpyDeviceToHost);

    return(res);
}

int multiply(float *src,float r0,float d0)
{
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*NX/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    multiply_kernel<<<grid,threads>>>(src,r0,d0,NZ*NY*NX);

    return(1);
}

int calc_pde(float2 *rho,float2 *fnl,float2 *pde)
{
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*PN/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    pde_kernel<<<grid,threads>>>(rho,fnl,pde,NZ*NY*PN);

    return(1);
}

int partial(float2 *fk,int dir)
{
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*PN/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    partial_kernel<<<grid,threads>>>(fk,dir,NZ*NY*PN);

    return(1);
}

int partial_add(float2 *in1,float2 *in2,int dir)
{
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*PN/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    partial_add_kernel<<<grid,threads>>>(in1,in2,dir,NZ*NY*PN);

    return(1);
}

int init_rhs(float2 *rhs)
{
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*PN/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    null_kernel<<<grid,threads>>>(rhs,NZ*NY*PN);

    return(1);
}

int update_rhok(float2 *rho,float2 *rhs)
{
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*PN/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    update_kernel<<<grid,threads>>>(rho,rhs,NZ*NY*PN);

    return(1);
}

int convolution(void)
{
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*PN/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    convolution_kernel<<<grid,threads>>>(rho_k,fnl_k,NZ*NY*PN);

    return(1);
}

int fdensity(void)
{
    dim3 threads(NTHREADS);
    int blocks=(int)sqrtf(NZ*NY*NX/NTHREADS)+1;
    dim3 grid(blocks,blocks);
    fdensity_kernel<<<grid,threads>>>(fnl_r,rho_r,NZ*NY*NX);

    return(1);
}

// *** GPU KERNELS ***
__global__ void norm_kernel(cufftComplex *data,int size,float factor)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   float2 v=data[idx];
   v.x*=factor;
   v.y*=factor;
   data[idx]=v;
 }
}

__global__ void filter_kernel(cufftComplex *data,int size,float factor)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   int tmp=idx;
   int k=tmp/(NY*PN);
   tmp-=k*NY*PN;
   int j=tmp/PN;
   tmp-=j*PN;
   int i=tmp;

   float kz=2.0f*PI/(NZ*HH)*(k < NZ/2+1 ? k : k-NZ);
   float ky=2.0f*PI/(NY*HH)*(j < NY/2+1 ? j : j-NY);
   float kx=2.0f*PI/(NX*HH)*i;
   float k2=kz*kz+ky*ky+kx*kx;

   int cond = ( k2 < KMAX*KMAX );

   float2 v=data[idx];
   v.x*=cond*factor;
   v.y*=cond*factor;
   data[idx]=v;
 }
}

__global__ void nonlin_kernel(float *src,float *dst,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   float v=src[idx];
   dst[idx]=NLFUNC(v);
 }
}

__global__ void multiply_kernel(float *fnl,float r0,float d0,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   int tmp=idx;
   int k=tmp/(NY*NX);
   tmp-=k*NY*NX;
   int j=tmp/NX;
   tmp-=j*NX;
   int i=tmp;

   float x=(i-NX/2)*HH;
   float y=(j-NY/2)*HH;
   float z=(k-NZ/2)*HH;
   float r=sqrtf(x*x+y*y+z*z);
   fnl[idx]*=MOBFNC(r,r0,d0);
 }
}

__global__ void pde_kernel(float2 *rho,float2 *fnl,float2 *pde,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   int tmp=idx;
   int k=tmp/(NY*PN);
   tmp-=k*NY*PN;
   int j=tmp/PN;
   tmp-=j*PN;
   int i=tmp;

   float kz=2.0f*PI/(NZ*HH)*(k < NZ/2+1 ? k : k-NZ);
   float ky=2.0f*PI/(NY*HH)*(j < NY/2+1 ? j : j-NY);
   float kx=2.0f*PI/(NX*HH)*i;
   float k2=kz*kz+ky*ky+kx*kx;

   float2 rho_t=rho[idx];
   float2 fnl_t=fnl[idx];

   float2 res;
   float Lk=LINOP(k2);
   res.x =Lk*rho_t.x+fnl_t.x;
   res.y =Lk*rho_t.y+fnl_t.y;

   pde[idx]=res;
 }
}

__global__ void update_kernel(float2 *rho,float2 *rhs,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   int tmp=idx;
   int k=tmp/(NY*PN);
   tmp-=k*NY*PN;
   int j=tmp/PN;
   tmp-=j*PN;
   int i=tmp;

   float kz=2.0f*PI/(NZ*HH)*(k < NZ/2+1 ? k : k-NZ);
   float ky=2.0f*PI/(NY*HH)*(j < NY/2+1 ? j : j-NY);
   float kx=2.0f*PI/(NX*HH)*i;
   float k2=kz*kz+ky*ky+kx*kx;

   float sop=1.0f+DT*k2*(LINOP(k2)+1.0f);
   float2 rho_t=rho[idx];
   float2 rhs_t=rhs[idx];

   float2 res;
   res.x =rho_t.x+DT/sop*rhs_t.x;
   res.y =rho_t.y+DT/sop*rhs_t.y;

   rho[idx]=res;
 }
}

__global__ void convolution_kernel(float2 *rho,float2 *fnl,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   int tmp=idx;
   int k=tmp/(NY*PN);
   tmp-=k*NY*PN;
   int j=tmp/PN;
   tmp-=j*PN;
   int i=tmp;

   float kz=2.0f*PI/(NZ*HH)*(k < NZ/2+1 ? k : k-NZ);
   float ky=2.0f*PI/(NY*HH)*(j < NY/2+1 ? j : j-NY);
   float kx=2.0f*PI/(NX*HH)*i;
   float k2=kz*kz+ky*ky+kx*kx;

   float Ck=LINOP(k2);
   float2 rho_t = rho[idx];
   float2 res;
   res.x = Ck*rho_t.x;
   res.y = Ck*rho_t.y;

   fnl[idx]=res;
 }
}

__global__ void partial_kernel(float2 *fk,int dir,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   int tmp=idx;
   int k=tmp/(NY*PN);
   tmp-=k*NY*PN;
   int j=tmp/PN;
   tmp-=j*PN;
   int i=tmp;

   float kz=2.0f*PI/(NZ*HH)*(k < NZ/2+1 ? k : k-NZ);
   float ky=2.0f*PI/(NY*HH)*(j < NY/2+1 ? j : j-NY);
   float kx=2.0f*PI/(NX*HH)*i;
   float k2=kz*kz+ky*ky+kx*kx;

   float w=0.0f;
   switch (dir)
   {
     case DIRX : { w=kx; } break;
     case DIRY : { w=ky; } break;
     case DIRZ : { w=kz; } break;
   }

   float2 f = fk[idx];
   float2 res;
   res.x = -w*f.y;
   res.y = +w*f.x;

   fk[idx]=res;
 }
}

__global__ void partial_add_kernel(float2 *in1,float2 *in2,int dir,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   int tmp=idx;
   int k=tmp/(NY*PN);
   tmp-=k*NY*PN;
   int j=tmp/PN;
   tmp-=j*PN;
   int i=tmp;

   float kz=2.0f*PI/(NZ*HH)*(k < NZ/2+1 ? k : k-NZ);
   float ky=2.0f*PI/(NY*HH)*(j < NY/2+1 ? j : j-NY);
   float kx=2.0f*PI/(NX*HH)*i;
   float k2=kz*kz+ky*ky+kx*kx;

   float w=0.0f;
   switch (dir)
   {
     case DIRX : { w=kx; } break;
     case DIRY : { w=ky; } break;
     case DIRZ : { w=kz; } break;
   }

   float2 out= in1[idx];
   float2 fk = in2[idx];
   float2 res;
   res.x = out.x-w*fk.y;
   res.y = out.y+w*fk.x;

   in1[idx]=res;
 }
}

__global__ void null_kernel(float2 *fk,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   float2 res;

   res.x=0.0f;
   res.y=0.0f;

   fk[idx]=res;
 }
}

__global__ void fdensity_kernel(float *csr,float *rho,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   float rho_t=rho[idx];
   float csr_t=csr[idx];
   csr[idx]=NLENRG(rho_t)+0.5f*rho_t*csr_t;
 }
}

__global__ void compare_kernel(float *src,float *ref,int size)
{
 int idx=(blockIdx.y*gridDim.x+blockIdx.x)*blockDim.x+threadIdx.x;

 if (idx<size)
 {
   float rho1=src[idx];
   float rho2=ref[idx];
   ref[idx]=(rho1-rho2)*(rho1-rho2);
 }
}
